
/**
 * Trida Stav representuje uzel grafu
 * 
 * @author Jiri Kremser 
 * @version 1.0    16. 10. 2006
 */

public class Stav{
  private int kanibal;  //pocet kanibalu
  private int misionar; //pocet misionaru
  private boolean breh; //kde je lodka
  public static boolean[][][] navstivene = new boolean[4][4][2]; //pole s priznaky navstivenych uzlu
  
  
  	/**
	 * Konstruktor tridy Stav
	 * 
	 * @param kan   pocet kanibalu
	 * @param mis   pocet misionaru
	 * @param br    kde je lodka
	 */
    public Stav(int kan,int mis,boolean br){
      this.kanibal=kan;
      this.misionar=mis;
      this.breh=br;
    }
    

    /**
	 * Prevede stav do textove podoby
	 * 
	 * @return     textova podoba stavu
	 */
    public String toStr(){
      int ka=3-kanibal;
      int mi=3-misionar;
      return " ["+kanibal+","+misionar+"]   |"+(breh?"*  |   [":"  *|   [")+ka+","+mi+"]";
    }
    
    
    /**
	 * Zjisti, zda uz byl stav navstiven
	 * 
	 * @param kan    pocet kanibalu testovaneho uzlu
	 * @param mis    pocet misionaru testovaneho uzlu
	 * @param br     stav lodky testovaneho uzlu
	 * @return     true byl-li navstiven
	 */
    public static boolean jeNavstiven(int kan, int mis, boolean br){
      if ( (kan>3) || (kan<0) || (mis>3) || (mis<0) ) {
         return true; 
      }
      else{
        for(int i=0;i<18;i++){
          if (navstivene[kan][mis][br?1:0]) return true;
        }
      }
      return false;
    }
     
    
    /**
	 * Zjisti, je-li stav konecny
	 * 
	 * @return     true je-li stav konecny
	 */    
    public boolean isEnd(){
      if ( (kanibal==0) && (misionar==0) && (!breh) ) return true;
      else return false;
    }
    
    
    /**
	 * Zjisti, je-li stav pripustny
	 * (pripustny ve smyslu pravidel daneho problemu, tj. vic misionaru nez kanibalu na brehu)
	 * 
	 * @return     true je-li stav pripustny
	 */       
    public static boolean jePripustny(int kan,int mis){
      if ( ((mis==3) && (kan==3)) || ((mis==3) && (kan==2)) || ((mis==3) && (kan==1))
        || ((mis==3) && (kan==0)) || ((mis==0) && (kan==3)) || ((mis==2) && (kan==2)) || ((mis==1) && (kan==1))
        || ((mis==0) && (kan==2)) || ((mis==0) && (kan==1)) || ((mis==0) && (kan==0)) )
        return true;
      else return false;
    }
    
    
    /**
	 * Nalezne mozne smysluplne operace nad danym uzlem/stavem
	 * (op1 = 2^0, op2 = 2^1, op3 = 2^2, op4 = 2^3, op5 = 2^4)
	 * 
	 * @return     soucet vsech moznych operatoru
	 */       
    public int mozneMoznosti(){
      int moznosti=0;
      
      if (!breh){
        if ( (!Stav.jeNavstiven(kanibal,misionar+1,!breh)) && (Stav.jePripustny(kanibal,misionar+1))) moznosti+=1;
        if ( (!Stav.jeNavstiven(kanibal,misionar+2,!breh)) && (Stav.jePripustny(kanibal,misionar+2))) moznosti+=2;
        if ( (!Stav.jeNavstiven(kanibal+1,misionar,!breh)) && (Stav.jePripustny(kanibal+1,misionar))) moznosti+=4;
        if ( (!Stav.jeNavstiven(kanibal+2,misionar,!breh)) && (Stav.jePripustny(kanibal+2,misionar))) moznosti+=8;
        if ( (!Stav.jeNavstiven(kanibal+1,misionar+1,!breh)) && (Stav.jePripustny(kanibal+1,misionar+1))) moznosti+=16;
      }
      else {
        if ( (!Stav.jeNavstiven(kanibal,misionar-1,!breh)) && (Stav.jePripustny(kanibal,misionar-1))) moznosti+=1;
        if ( (!Stav.jeNavstiven(kanibal,misionar-2,!breh)) && (Stav.jePripustny(kanibal,misionar-2))) moznosti+=2;
        if ( (!Stav.jeNavstiven(kanibal-1,misionar,!breh)) && (Stav.jePripustny(kanibal-1,misionar))) moznosti+=4;
        if ( (!Stav.jeNavstiven(kanibal-2,misionar,!breh)) && (Stav.jePripustny(kanibal-2,misionar))) moznosti+=8;
        if ( (!Stav.jeNavstiven(kanibal-1,misionar-1,!breh)) && (Stav.jePripustny(kanibal-1,misionar-1))) moznosti+=16;
      }
      return moznosti;
    }
    
    
    /**
	 * Realizace operace nad danym stavem / prechod z uzlu o uroven niz
	 *  
	 * @param kod  binarni soucet ntych mocnin dvojky, kde n je kod operace
	 * @return     potomek/poduzel
	 */   
    public Stav operace(int kod){
      Stav zmeneny=new Stav(kanibal,misionar,breh);    
      switch (kod){
        case 1:
          if (breh) zmeneny.misionar=misionar-1;
          else zmeneny.misionar=misionar+1;
        break;
        case 2:
          if (breh) zmeneny.misionar=misionar-2;
          else zmeneny.misionar=misionar+2;
        break;
        case 3:
          if (breh) zmeneny.kanibal=kanibal-1;
          else zmeneny.kanibal=kanibal+1;
        break;
        case 4:
          if (breh) zmeneny.kanibal-=2;
          else zmeneny.kanibal=kanibal+2;
        break;
        case 5:
          if (breh) {
              zmeneny.misionar=misionar-1;
              zmeneny.kanibal=kanibal-1;
          }
          else {
            zmeneny.misionar=misionar+1;
            zmeneny.kanibal=kanibal+1;
          }
          
        break;
      }
      zmeneny.breh = !zmeneny.breh;
  
      navstivene[zmeneny.kanibal][zmeneny.misionar][zmeneny.breh?1:0]=true;
      return zmeneny;
    }
    
    
    /**
	 * Spocita a vypise reseni problemu z daneho stavu
	 */   
    public void pocitej(){
      boolean first=true;
      Stav stavek = Vypocet.q.dequeue();
      
      while (!stavek.isEnd()||first){
        if (!first) stavek = Vypocet.q.dequeue();
        first=false;

        while (!Stav.jeNavstiven(stavek.kanibal,stavek.misionar,stavek.breh) && !Vypocet.q.isEmpty()) stavek = Vypocet.q.dequeue();
        if (stavek.mozneMoznosti()!=0) System.out.println(stavek.toStr());
        if (stavek.isEnd()) System.out.println("KONEC");
        // projede a zaradi do fronty stavy, do nez lze vstoupit
        if ((stavek.mozneMoznosti()&1)==1) Vypocet.q.enqueue(stavek.operace(1));
        if ((stavek.mozneMoznosti()&2)==2) Vypocet.q.enqueue(stavek.operace(2));
        if ((stavek.mozneMoznosti()&4)==4) Vypocet.q.enqueue(stavek.operace(3));
        if ((stavek.mozneMoznosti()&8)==8) Vypocet.q.enqueue(stavek.operace(4));
        if ((stavek.mozneMoznosti()&16)==16) Vypocet.q.enqueue(stavek.operace(5));
      }
    }
    
}